#!/bin/sh
#
# build.sh - build script for Cetus.
#

# Modify these variables for your system.
ANTLR="$PWD/lib/antlr.jar"            #antlr location

if [ -z "$ANTLR" ]; then
  echo "Please define ANTLR in $0"
  exit 1
fi

# Check for java/javac/jar.
for tool in java javac jar; do
  which $tool >/dev/null
  if [ $? -ne 0 ]; then
    echo $tool is not found.
    exit 1
  fi
done

# No change is required for these variables.
CETUSROOT=$PWD
SRC="$CETUSROOT/src/*/*/*.java $CETUSROOT/src/*/*/*/*.java"
PARSER="$CETUSROOT/src/cetus/base/grammars"
# Source files for parser construction.
parser_src="
  CToken.java
  LineObject.java
  NewCParser.g
  Pragma.java
  Pre.g
  PreprocessorInfoChannel.java
	CetusCParser.java
"

case "$1" in
  parser)
  echo "Compiling the parser using ANTLR..."
  #Build the parser
  cd $PARSER
  java -cp $ANTLR:class antlr.Tool $PARSER/Pre.g
  java -cp $ANTLR:class antlr.Tool $PARSER/NewCParser.g
  cd -
  ;;
  compile)
  echo "Compiling the source files..."
  if [ ! -f $PARSER/NewCParser.java ]; then
    $0 parser
  fi
  [ -d class ] || mkdir class
  javac -g -cp $ANTLR:class -d class $SRC
  ;;
  jar)
  $0 compile
  echo "Archiving the class files..."
  [ -d lib ] || mkdir lib
  jar cf lib/cetus.jar -C class .
  ;;
  javadoc)
  echo "Generating JAVA documents..."
  javadoc -d api cetus $SRC
  ;;
  manual)
  echo "Generating Cetus manual..."
  cd manual
  docbook2html -o html manual.xml
  docbook2html -u -o html manual.xml
  docbook2pdf manual.xml
  cd -
  ;;
  clean)
  echo "Cleaning up..."
  rm -rf class bin lib/cetus.jar
  ;;
  purge)
  $0 clean
  echo "Purging classes and documents..."
  # removes files under grammars directory except for the source files.
  cd $PARSER; mv $parser_src ..; rm -f *
  cd ..; mv $parser_src grammars/
  cd $CETUSROOT
  rm -rf api manual/html manual/manual.pdf
  ;;
  bin)
  $0 jar
  echo "Generating a wrapper..."
  [ -d bin ] || mkdir bin
  classpath="$ANTLR:$PWD/lib/cetus.jar"
  cat > bin/cetus << EOF
#!/bin/sh
# This file was generated automatically by build.sh.
java -cp $classpath -Xms500m -Xmx500m cetus.exec.Driver \$*
EOF
  chmod 755 bin/cetus
  ;;
  *)
  echo "Usage: $0 <target>"
  echo "  <target> is one of the followings:"
  echo "  bin     - compile, jar, and generate a wrapper script."
  echo "  compile - compile the source files."
  echo "  clean   - remove classes, jar, and the wrapper."
  echo "  jar     - archive the class files."
  echo "  javadoc - generate api documents."
  echo "  parser  - rebuild the parser."
  echo "  purge   - remove classes, jar, wrapper, and documents."
  exit 1
  ;;
esac
