package benchmark;

import cetus.exec.Driver;
import cetus.exec.CommandLineOptionSet;
import cetus.hir.Program;
import java.util.List;
import java.util.ArrayList;

/**
* Driver that exposes option sets and IR to public.
* The main purpose of providing this derived class is to allow accesses to
* objects such as option set and program IR which are protected members within
* a standard driver (cetus.exec.Driver).
*/
public class CetusDriver extends cetus.exec.Driver {

    /**
    * Constructs a new cetus driver with default options.
    */
    public CetusDriver() {
        super();
        setDefaultOptions();
    }

    /**
    * Modifies the cetus option with the given key and value.
    * @param key the name of the cetus option to be modified
    * @param value the new value of the <b>key</b>
    */
    public void setOption(String key, String value) {
        options.setValue(key, value);
    }

    /**
    * Returns the string value of the cetus option having the given name.
    * @param key the name of the cetus option
    * @return the value of the ceuts option or null if no such option found
    */
    public String getOption(String key) {
        return options.getValue(key);
    }

    /**
    * Attaches the specified <b>program</b> to the current driver.
    * This method can be used in a case where a single master program IR is
    * used with different drivers over and over while skipping individual
    * parsing steps.
    */
    public void setProgram(Program program) {
        this.program = program;
    }

    /**
    * Returns the program IR attached to the current driver.
    * @return the program IR
    */
    public Program getProgram() {
        return program;
    }

    /**
    * Parses the files with the specified names and returns the resulting
    * program IR.
    * This method can be used when the caller of this method ensures the
    * current working directory is identical to the directory containing the
    * input files.
    * @param files the names of the input C programs.
    * @return the parsed program IR.
    */
    public Program parse(String[] files) {
        filenames = new ArrayList<String>(files.length);
        for (String file : files) {
            filenames.add(file);
        }
        parseFiles();
        return program;
    }

    /**
    * Parses the files with the specified names after changing the current
    * working directory to the one specified by <b>dir</b>.
    * @param files the names of the input C programs.
    * @return the parsed program IR.
    */
    public Program parse(String[] files, String dir) {
        String user_dir = System.getProperty("user.dir");
        System.setProperty("user.dir", dir);
        parse(files);
        System.setProperty("user.dir", user_dir);
        return program;
    }

    /**
    * Filles the option set with some default values.
    */
    private void setDefaultOptions() {
        options.setValue("outdir", "cetus_output");
        options.setValue("preprocessor", "cpp -C -I.");
        options.setValue("verbosity", "0");
        options.setValue("alias", "1");
    }

}
