package benchmark;

import java.util.Map;
import java.util.LinkedHashMap;
import java.lang.Number;
import cetus.hir.PrintTools;

/**
* This class provides container for collecting metrics from a benchmark
* invocation.
* <p>
* The stored data are simply set of triples, each of which consists
* of (NAME, ID, NUMBER):
* <ul>
* <li> <b>NAME</b>: the name of the metric
* <li> <b>ID</b>: the string that identifies a certain code section or IR
* <li> <b>NUMBER</b>: the metric value expressed as a number
* </ul>
* At this point, no restriction is applied to each element of this triple,
* but commonly used metrics will be defined later.
*/
public class Metric {

    /** Metric entries. */
    private Map<String, Map<String, Number>> entries;

    /** Constructs a new metric object. */
    public Metric() {
        entries = new LinkedHashMap<String, Map<String, Number>>();
    }

    /**
    * Adds the given arguments to the metric object.
    * The value argument is converted to a number internally, following the
    * definition of a valid metric entry.
    * @param name the name of the metric
    * @param id the identifying string for the given metric name
    * @param value the value of the metric entry
    */
    public void add(String name, String id, String value) {
        Number number;
        try {
            number = new Long(value);
        } catch (NumberFormatException e) {
            try {
                number = new Double(value);
            } catch (NumberFormatException ex) {
                throw new IllegalArgumentException();
            }
        }
        add(name, id, number);
    }

    /**
    * Adds the given arguments to the metric object.
    * @param name the name of the metric
    * @param id the identifying string within the given metric name
    * @param value the value of the metric entry
    */
    public void add(String name, String id, Number value) {
        Map<String, Number> entry = entries.get(name);
        if (entry == null) {
            entry = new LinkedHashMap<String, Number>();
            entries.put(name, entry);
        }
        entry.put(id, value);
    }

    /**
    * Returns the metric entries associated with the given metric name,
    * <b>name</b>, as a map.
    * @param name the name of the metric
    * @return the map from metric identifier to its value
    */
    public Map<String, Number> getEntries(String name) {
        return entries.get(name);
    }

    /**
    * Returns the metric value associated with the given name and identifier.
    * @param name the name of the metric
    * @param id the identifying string within the given metric name
    * @return the value associated or null (if not found)
    */
    public Number get(String name, String id) {
        Map<String, Number> entry = entries.get(name);
        if (entry == null) {
            return null;
        } else {
            return entry.get(id);
        }
    }

    /**
    * Returns a raw string that lists the contents of the metric.
    * @return the string
    */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (String name : entries.keySet()) {
            Map<String, Number> entry = entries.get(name);
            for (String id : entry.keySet()) {
                Number num = entry.get(id);
                sb.append(name).append(" ").append(id).append(" ").append(num);
                sb.append(PrintTools.line_sep);
            }
        }
        return sb.toString();
    }

}
