package benchmark;

import java.util.Map;
import java.util.TreeMap;
import java.util.List;
import java.util.LinkedList;
import java.util.Arrays;

/**
* Class {@code Option} contains the parameters (to be) used for a specific run
* of a single benchmarking.
* <p>
* A single benchmarking consists of cetus translation,
* native code generation, and program execution. An instance of this class is
* equivalent to a specific benchmark run and can be created through API or
* from a file in Python's configuration file format. The following template
* completely explains the currently available variables within an Option object
* and provides a template for an option file. This template can also be
* generated by calling {@link #getUsage()}.
* <p>
* <pre>
; Comment starts with either ';' or '#'.
; OPTIONAL: the variable does not need to be defined
; REQUIRED: the variable needs to be defined
; LISTABLE: the variable may contain multiple values separated by a line separator and at least one white space (or tab)
; APPENDABLE: the variable's value cumulativley appended with a separating white space

[cetus]
exclude =  
    ; List of files to be excluded during Cetus translation
    ; OPTIONAL, LISTABLE
extraoption =  
    ; List of extra options required for a specific program
    ; OPTIONAL, LISTABLE
option =  
    ; List of options for Cetus translation
    ; OPTIONAL, LISTABLE
task =  
    ; List of tasks to be performed before Cetus translation
    ; OPTIONAL, LISTABLE

[common]
cpp =  
    ; The C preprocessor command
    ; OPTIONAL, DEFAULT = cpp
cppflags =  
    ; C preprocessor flags
    ; OPTIONAL, APPENDABLE
dir =  
    ; Location of the root benchmark directory
    ; OPTIONAL, DEFAULT = current working directory
env =  
    ; List of environment variable definitions
    ; OPTIONAL, LISTABLE
name =  
    ; Name of the benchmark instance
    ; OPTIONAL, DEFAULT = automatic name based on current time
src =  
    ; Location of the C source code directory
    ; REQUIRED

[compile]
compiler =  
    ; C compiler command
    ; OPTIONAL, DEFAULT = gcc
extraflags =  
    ; Extra C flags for a specific program
    ; OPTIONAL
optimize =  
    ; C optimization flags
    ; OPTIONAL, DEFAULT = -O1
task =  
    ; List of tasks to be performed for program compilation
    ; REQUIRED

[run]
task =  
    ; List of tasks to be performed for program execution
    ; REQUIRED
validate =  
    ; List of tasks to be performed for program validation
    ; OPTIONAL
validon =  
    ; String to be matched for validation in the stdout file
    ; OPTIONAL
* </pre>
* <p>
* Here is a real example option file defined for a SPEC CPU2006 400.perlbench:
* <pre>
[common]
src      = src
; 64 bit 
cppflags = -I. -DSPEC_CPU -DPERL_CORE -DSPEC_CPU_LINUX_X64 -DSPEC_CPU_LP64
; 32 bit
;cppflags = -I. -DSPEC_CPU -DPERL_CORE -DSPEC_CPU_LINUX_IA32
env      = SPEC=../../..

[cetus]
exclude     = hparser.c
              parser-util.c
extraoption = expand-all-header=1

[compile]
task = make

[run]
task     = cp ../compile/perlbench .
           cp -r ../../data-test/in/* .
           $SPEC/bin/specinvoke -E speccmds.cmd
validate = $SPEC/bin/specinvoke -E compare.cmd
* </pre>
* <p>
* One invariant implied within an option file (or object) is that the benchmark
* invocation process (included in the <b>benchmarktool</b> Jython module) starts
* in the directory specified with <b>[common] dir =</b> and creates an instance
* directory specified with <b>[common] name =</b> to store files created during
* a benchmark invocation. Following actions summarizes the semantics of an
* option file (or object) in sequential order:
* <p>
* <ul>
* <li> Creates an invocation directory (will call <b>INVDIR</b>)
* <li> Copies the entire source code directory to <b>INVDIR</b>
* <li> Performs tasks defined within the [cetus] section if applicable
* <li> Creates directory named <b>compile</b> within <b>INVDIR</b>
* <li> Invokes Cetus driver and places the result within <b>INVDIR/compile</b>
* <li> Performs the compilation tasks as defined within the <b>[compile]</b>
*      section in <b>INVDIR/compile</b>
* <li> Performs the run tasks as defined within the <b>[run]</b> section in
*      <b>INVDIR/run</b>
* </ul>
*/
public class Option implements Cloneable {

    /**
    * Type definitions for each option.
    * Three types are supported:
    */
    private static Map<String, Integer> types;
    private static final int ELEMENTARY = 0;
    private static final int LISTABLE   = 1;
    private static final int APPENDABLE = 2;

    /**
    * Simple patterns that defines option values which represents a map itself.
    * The splitting string is stored for those options.
    */
    private static Map<String, String> splitters;

    /** Template option that shows all the allowed entries */
    private static Option document;

    static {
        defineOption();
    }

    /**
    * Defines the option entries by assigning their types, patterns, and
    * semantics.
    */
    private static void defineOption() {
        types = new TreeMap<String, Integer>();
        splitters = new TreeMap<String, String>();
        document = new Option();
        // common section
        types.put("common" + ":" + "src", ELEMENTARY);
        types.put("common" + ":" + "dir", ELEMENTARY);
        types.put("common" + ":" + "name", ELEMENTARY);
        types.put("common" + ":" + "cpp", ELEMENTARY);
        types.put("common" + ":" + "env", LISTABLE);
        types.put("common" + ":" + "cppflags", APPENDABLE);
        Map<String, List<String>> section = new TreeMap<String, List<String>>();
        document.options.put("common", section);
        section.put("src", Arrays.asList(" ",
                "; Location of the C source code directory",
                "; REQUIRED"));
        section.put("dir", Arrays.asList(" ",
                "; Location of the root benchmark directory",
                "; OPTIONAL, DEFAULT = current working directory"));
        section.put("name", Arrays.asList(" ",
                "; Name of the benchmark instance",
                "; OPTIONAL, DEFAULT = automatic name based on current time"));
        section.put("cpp", Arrays.asList(" ",
                "; The C preprocessor command",
                "; OPTIONAL, DEFAULT = cpp"));
        section.put("env", Arrays.asList(" ",
                "; List of environment variable definitions",
                "; OPTIONAL, LISTABLE"));
        section.put("cppflags", Arrays.asList(" ",
                "; C preprocessor flags",
                "; OPTIONAL, APPENDABLE"));
        splitters.put("common" + ":" + "env", "=");
        // cetus section
        types.put("cetus" + ":" + "exclude", LISTABLE);
        types.put("cetus" + ":" + "option", LISTABLE);
        types.put("cetus" + ":" + "extraoption", LISTABLE);
        types.put("cetus" + ":" + "task", LISTABLE);
        section = new TreeMap<String, List<String>>();
        document.options.put("cetus", section);
        section.put("exclude", Arrays.asList(" ",
                "; List of files to be excluded during Cetus translation",
                "; OPTIONAL, LISTABLE"));
        section.put("option", Arrays.asList(" ",
                "; List of options for Cetus translation",
                "; OPTIONAL, LISTABLE"));
        section.put("extraoption", Arrays.asList(" ",
                "; List of extra options required for a specific program",
                "; OPTIONAL, LISTABLE"));
        section.put("task", Arrays.asList(" ",
                "; List of tasks to be performed before Cetus translation",
                "; OPTIONAL, LISTABLE"));
        splitters.put("cetus" + ":" + "option", "=");
        splitters.put("cetus" + ":" + "extraoption", "=");
        // compile section
        types.put("compile" + ":" + "compiler", ELEMENTARY);
        types.put("compile" + ":" + "optimize", APPENDABLE);
        types.put("compile" + ":" + "extraflags", APPENDABLE);
        types.put("compile" + ":" + "task", LISTABLE);
        section = new TreeMap<String, List<String>>();
        document.options.put("compile", section);
        section.put("compiler", Arrays.asList(" ",
                "; C compiler command",
                "; OPTIONAL, DEFAULT = gcc"));
        section.put("optimize", Arrays.asList(" ",
                "; C optimization flags",
                "; OPTIONAL, DEFAULT = -O1"));
        section.put("extraflags", Arrays.asList(" ",
                "; Extra C flags for a specific program",
                "; OPTIONAL"));
        section.put("task", Arrays.asList(" ",
                "; List of tasks to be performed for program compilation",
                "; REQUIRED"));
        // run section
        types.put("run" + ":" + "task", LISTABLE);
        types.put("run" + ":" + "validate", LISTABLE);
        types.put("run" + ":" + "validon", ELEMENTARY);
        section = new TreeMap<String, List<String>>();
        document.options.put("run", section);
        section.put("task", Arrays.asList(" ",
                "; List of tasks to be performed for program execution",
                "; REQUIRED"));
        section.put("validate", Arrays.asList(" ",
                "; List of tasks to be performed for program validation",
                "; OPTIONAL"));
        section.put("validon", Arrays.asList(" ",
                "; String to be matched for validation in the stdout file",
                "; OPTIONAL"));
    }

    private int getType(String section, String key) {
        Integer ret = types.get(section + ":" + key);
        if (ret == null) {
            ret = ELEMENTARY;
        }
        return ret;
    }

    /**
    * Options directly converted from python configuration.
    * A configuration description from python consists of three string values;
    * configuration section, configuration name, and configuration value.
    */
    private Map<String, Map<String, List<String>>> options;

    /** Constructs a new instance with automatically */
    public Option() {
        options = new TreeMap<String, Map<String, List<String>>>();
    }

    /**
    * Sets the specified key and value within the given section.
    * Already existing option is overwritten by the new key and value.
    * @param section the name of the section
    * @param key the key string (option name) to be set
    * @param value the value string to be set for the option
    */
    public void set(String section, String key, String value) {
        Map<String, List<String>> s = options.get(section);
        if (s == null) {
            s = new TreeMap<String, List<String>>();
            options.put(section, s);
        }
        List<String> values = new LinkedList<String>();
        values.add(value);
        s.put(key, values);
    }

    /**
    * Appends the specified value for the given section and key.
    * Unlike {@link #set(String, String, String)}, this method modifies an
    * existing value if applicable. If the key does not exist, the behavior of
    * this method is identical to that of {@link #set(String, String, String)}.
    * @param section the name of the section
    * @param key the key string (option name) to be set
    * @param value the value string to be set for the option
    */
    public void append(String section, String key, String value) {
        int type = getType(section, key);
        if (!contains(section, key)) {
            set(section, key, value);
        } else if (type == ELEMENTARY) {
            set(section, key, value);
        } else {
            List<String> values = options.get(section).get(key);
            if (type == LISTABLE) {
                values.add(value);
            } else if (type == APPENDABLE) {
                values.set(0, values.get(0) + " " + value);
            }
        }
    }

    /**
    * Removes the option associated with the given <b>section</b> and
    * <b>key</b>.
    * @param section the name of the section
    * @param key the key string for the option name
    * @return the value of the option being removed, or null if no such option
    *       exists.
    */
    public void remove(String section, String key) {
        if (contains(section, key)) {
            options.get(section).remove(key);
            if (options.get(section).isEmpty()) {
                options.remove(section);
            }
        }
    }

    /**
    * Returns the option value associated with <b>section</b> and <b>key</b>.
    * @param section the name of the section
    * @param key the key string for the option name
    * @return the value of the option, or null if no such option exists or
    *       the key is recognized as a listable option
    */
    public String get(String section, String key) {
        if (!contains(section, key)) {
            return null;
        }
        if (getType(section, key) == LISTABLE) {
            System.err.println(section + ":" + key +
                    " is recognized as a listable option");
            return null;
        }
        return options.get(section).get(key).get(0);
    }

    /**
    * Returns the list of option values associated with <b>section</b> and
    * <b>key</b>.
    * @param section the name of the section
    * @param key the key string for the option name
    * @return the value of the option, or null if no such option exists or
    *       the key is not recognized as a listable option
    */
    public List<String> getList(String section, String key) {
        if (!contains(section, key)) {
            return new LinkedList<String>();
        }
        if (getType(section, key) != LISTABLE) {
            System.err.println(section + ":" + key +
                    " is not recognized as a listable option");
            return new LinkedList<String>();
        }
        return options.get(section).get(key);
    }

    /**
    * Returns a map constructed from the specified <b>section</b> and
    * <b>key</b>.
    * If the specified option value is predefined as having a splitting string
    * that represents some type of name and its value, this method extracts a
    * map from the option value and returns the resulting map.
    * @param section the name of the section
    * @param key the key string for the option name
    * @return the resulting map
    */
    public Map<String, String> getMap(String section, String key) {
        Map<String, String> ret = new TreeMap<String, String>();
        if (!contains(section, key)) {
            return ret;
        }
        String k = section + ":" + key;
        if (!splitters.containsKey(k)) {
            System.err.println(k + " is not recognized as a splittable option");
            return ret;
        }
        List<String> values = options.get(section).get(key);
        String sp = splitters.get(k);
        for (String value : values) {
            int pos = value.indexOf(sp);
            if (pos == -1) {
                System.err.println(value + " is not splittable with " + sp);
            } else {
                String lhs = value.substring(0, pos).trim();
                String rhs = value.substring(pos+1).trim();
                if (lhs.length() == 0 || rhs.length() == 0) {
                    System.err.println(value + " is not splittable with " + sp);
                } else {
                    ret.put(lhs, rhs);
                }
            }
        }
        return ret;
    }

    /**
    * Checks if there is an option associated with <b>section</b> and
    * <b>key</b>.
    * @param section the name of the section
    * @param key the key string for the option name
    * @return true if one exists, false otherwise
    */
    public boolean contains(String section, String key) {
        Map<String, List<String>> s = options.get(section);
        if (s == null) {
            return false;
        } else {
            return s.containsKey(key);
        }
    }

    /**
    * Returns an example option file contents by listing all the recognized
    * options and their semantics.
    */
    public static String getUsage() {
        String sep = System.getProperty("line.separator");
        String ret = "";
        ret += "; OPTIONAL: the variable does not need to be defined" + sep;
        ret += "; REQUIRED: the variable needs to be defined" + sep;
        ret += "; LISTABLE: the variable may contain multiple values separated by a line separator and at least one white space (or tab)" + sep;
        ret += "; APPENDABLE: the variable's value cumulativley appended with a separating white space" + sep;
        return ret + document.dump();
    }

    private String dump() {
        StringBuilder sb = new StringBuilder();
        String sep = System.getProperty("line.separator");
        for (String section : options.keySet()) {
            sb.append("[").append(section).append("]").append(sep);
            for (String option : options.get(section).keySet()) {
                int type = getType(section, option);
                List<String> values = options.get(section).get(option);
                int indent = 4;
                sb.append(option).append(" = ").append(values.get(0));
                for (int i = 1; i < values.size(); i++) {
                    sb.append(sep);
                    for (int j = 0; j < indent; j++) {
                        sb.append(" ");
                    }
                    sb.append(values.get(i));
                }
                sb.append(sep);
            }
            sb.append(sep);
        }
        return sb.toString();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        String sep = System.getProperty("line.separator");
        for (String section : options.keySet()) {
            sb.append("[").append(section).append("]").append(sep);
            for (String option : options.get(section).keySet()) {
                int type = getType(section, option);
                List<String> values = options.get(section).get(option);
                if (type == LISTABLE) {
                    int indent = option.length() + 3;
                    sb.append(option).append(" = ").append(values.get(0));
                    for (int i = 1; i < values.size(); i++) {
                        sb.append(sep);
                        for (int j = 0; j < indent; j++) {
                            sb.append(" ");
                        }
                        sb.append(values.get(i));
                    }
                } else {
                    sb.append(option).append(" = ").append(values.get(0));
                }
                sb.append(sep);
            }
            sb.append(sep);
        }
        return sb.toString();
    }

    @Override
    public Option clone() {
        Option ret = new Option();
        for (String s : options.keySet()) {
            Map<String, List<String>> section =
                    new TreeMap<String, List<String>>();
            ret.options.put(s, section);
            for (String k : options.get(s).keySet()) {
                section.put(k, new LinkedList<String>(options.get(s).get(k)));
            }
        }
        return ret;
    }

}
