class Benchmark(object):
    '''
    Provides a module that performs processes defined in an option object.
    The following attributes are directly accessible for a Benchmark object.
        option      the associated option object
        metric      the associated metric object
    '''

    def __init__(self, option=None, optfile=None,
                 stdout=None, stderr=None, stdjoin=False):
        '''
        Constructs a new Benchmark object.
        If both 'option' and 'optfile' are passed, 'optfile' is discarded.
            option      option object
            optfile     option file name
            stdout      file descriptor to which stdout is redirected
            stderr      file descriptor to which stderr is redirected
            stdjoin     join stdout and stderr if this is true
        '''
        import os
        self.stdout = stdout
        self.stderr = stderr
        self.stdjoin = stdjoin
        self.option = option
        self.cwd = os.getcwd()
        self.envs = dict(os.environ)
        if not option and optfile:
            self.parse_option(optfile)

    def __info(self, msg):
        '''
        Prints information.
        '''
        import os
        import sys
        msg = '[%10s'%'INFO] ' + msg + os.linesep
        sys.__stderr__.write(msg)
        if self.stderr:
            self.stderr.write(msg)
        return

    def __error(self, msg):
        '''
        Prints error messages.
        '''
        import os
        import sys
        msg = '[%10s'%'ERROR] ' + msg + os.linesep
        sys.__stderr__.write(msg)
        if self.stderr:
            self.stderr.write(msg)
        return

    def __setup(self):
        '''
        Prepares a benchmark instance.
        This method performs the following tasks: 
            - Checks if the benchmark consists of valid options
            - Filles the option with default values
            - Prepares benchmark directory
            return      True if everything is fine
        '''
        from benchmark import Option
        from benchmark import Metric
        import os
        import time
        # Resolves absolute location of the benchmark directory.
        if self.option.contains('common', 'dir'):
            rootdir = self.option.get('common', 'dir')
            if not rootdir.startswith(os.sep):
                rootdir = os.path.join(self.cwd, rootdir)
        else:
            rootdir = self.cwd
        self.option.set('common', 'dir', rootdir)
        # Resolves absolute location of the source code directory.
        if not self.option.contains('common', 'src'):
            self.__error('Location of C source codes is missing.')
            return False
        src = self.option.get('common', 'src')
        if not src.startswith(os.sep):
            src = os.path.join(rootdir, src)
            self.option.set('common', 'src', src)
        # Checks if 'src' and 'rootdir' are accessible.
        if not os.access(rootdir, os.F_OK):
            self.__error('Cannot access the benchmark directory: ' + rootdir)
            return False
        if not os.access(src, os.F_OK):
            self.__error('Cannot access the source code directory: ' + src)
            return False
        # Automatically generate name of the benchmark if not defined.
        if not self.option.contains('common', 'name'):
            name = 'benchmark.' + time().strftime('%m%d%H%M%S')
            self.option.set('common', 'name', name)
        # Prepares the benchmark instance directory.
        self.benchdir = os.path.join(rootdir, self.option.get('common', 'name'))
        self.compiledir = os.path.join(self.benchdir, 'compile')
        self.rundir = os.path.join(self.benchdir, 'run')
        try:
            os.mkdir(self.benchdir)
            if not self.stdout:
                self.stdout = \
                        open(os.path.join(self.benchdir, 'benchmark.out'), 'w')
            if self.stdjoin:
                self.stderr = self.stdout
            elif not self.stderr:
                self.stderr = \
                        open(os.path.join(self.benchdir, 'benchmark.err'), 'w')
            if self.stdout == self.stderr and not self.stdjoin:
                self.stdjoin = True
        except Exception, e:
            self.__error('Failed to create a benchmark instance')
            self.__error(e)
            return False
        # Assigns metric object.
        self.metric = Metric()
        # Environment variables
        envs = self.option.getMap("common", "env")
        for env in envs:
            self.envs[env] = envs.get(env)
        self.envs['OMP_NUM_THREADS'] = self.option.get('run', 'threads')
        self.__info('Benchmark root: ' + rootdir)
        self.__info('Benchmark name: ' + self.option.get('common', 'name'))
        return True

    def __fill_defaults(self):
        '''
        Filles the option object with default values.
        '''
        self.option.set('common', 'cpp', 'cpp')
        self.option.set('common', 'cppflags', '-I.')
        self.option.set('compile', 'compiler', 'gcc')
        self.option.set('compile', 'optimize', '-O1')
        self.option.set('run', 'threads', '1')
        return

    def __invoke(self, task):
        '''
        Invokes the specified task string with optional metric collector.
            task        task string to be invoked
        '''
        import os
        import sys
        import subprocess
        # prepares temporary files
        outfile = os.path.join(os.getcwd(), 'invoke.out')
        errfile = os.path.join(os.getcwd(), 'invoke.err')
        if os.access(outfile, os.F_OK):
            os.remove(outfile)
        if os.access(errfile, os.F_OK):
            os.remove(errfile)
        out = open(outfile, 'w')
        err = open(errfile, 'w')
        try:
            retcode = subprocess.call(task, shell=True, stdout=out,
                                      stderr=err, env=self.envs)
            out.close()
            err.close()
            for line in open(outfile, 'r'):
                sys.stdout.write(line)
            for line in open(errfile, 'r'):
                sys.stderr.write(line)
            ret = True
        except Exception, e:
            if not out.closed:
                out.close()
            if not err.closed:
                err.close()
            self.__error('Failed to invoke a task: ' + task)
            self.__error(e)
            ret = False
            retcode = -1
        finally:
            os.remove(outfile)
            os.remove(errfile)
            return (ret, retcode)

    def __report(self, phase, result, runtime):
        self.metric.add(phase + '.pass', 'program', result)
        self.metric.add(phase + '.time', 'program', runtime)

    def __invoke_cetus(self, runpass=True, runprint=True):
        '''
        Invokes a cetus driver.
            runpass     whether or not it runs passes in the cetus driver
            return      (boolean result, Program IR)
        '''
        import os
        import shutil
        import fnmatch
        import time
        from benchmark import CetusDriver
        self.__info('Invoking a cetus driver.')
        os.chdir(self.benchdir)
        program = None
        t = time.time()
        # copies the whole source code directory to the benchmark directory
        try:
            shutil.copytree(self.option.get('common', 'src'), 'src')
        except Exception, e:
            self.__error('Failed to copy the source code directory.')
            self.__error(e)
            self.__report('cetus', 0, time.time() - t)
            return (False, program)
        # performs tasks defined for the cetus section.
        if self.option.contains('cetus', 'task'):
            for task in self.option.getList('cetus', 'task'):
                (ret0, ret1) = self.__invoke(task)
                if not ret0:
                    self.__report('cetus', 0, time.time() - t)
                    return (False, program)
        # copies the original source code to 'compile' directory.
        try:
            shutil.copytree('src', 'compile')
        except Exception, e:
            self.__error('Failed to create compile directory.')
            self.__error(e)
            self.__report('cetus', 0, time.time() - t)
            return (False, program)
        # constructs a cetus driver and sets options.
        os.chdir('src')
        cetus = CetusDriver()
        # preprocessor
        if self.option.contains('common', 'cpp'):
            cpp = self.option.get('common', 'cpp')
            if self.option.contains('common', 'cppflags'):
                cpp += ' ' + self.option.get('common', 'cppflags')
            cetus.setOption('preprocessor', cpp)
        # output directory
        cetus.setOption('outdir', self.compiledir)
        # user options
        if self.option.contains('cetus', 'option'):
            cetus_opts = self.option.getMap('cetus', 'option')
            for key in cetus_opts:
                cetus.setOption(key, cetus_opts.get(key))
        # portability options
        if self.option.contains('cetus', 'extraoption'):
            extra_opts = self.option.getMap('cetus', 'extraoption')
            for key in extra_opts:
                cetus.setOption(key, extra_opts.get(key))
        # creates source code list
        srcs = []
        srcdir = os.path.join(self.benchdir, 'src')
        excludes = self.option.getList('cetus', 'exclude')
        for f in os.listdir(srcdir):
            if fnmatch.fnmatch(f, '*.c') and not excludes.contains(f):
                srcs.append(os.path.join(srcdir, f))
        # inovkes cetus
        try:
            self.__start_logj()
            program = cetus.parse(srcs, srcdir)
            if runpass:
                cetus.runPasses()
            if runprint:
                program.print()
            self.__stop_logj()
        except Exception, e:
            self.__error('Failed to run cetus')
            self.__error(e)
            self.__report('cetus', 0, time.time() - t)
            return (False, program)
        self.__info('Finished cetus translation.')
        self.__report('cetus', 1, time.time() - t)
        return (True, program)

    def __close_fds(self):
        '''
        Closes the benchmark object.
        '''
        if self.stdout and not self.stdout.closed:
            self.stdout.close()
        if self.stderr and not self.stderr.closed:
            self.stderr.close()
        return

    def __invoke_compile(self):
        '''
        Invokes tasks defined for the compile section.
        '''
        import os
        import time
        self.__info('Compiling the program.')
        os.chdir(self.compiledir)
        t = time.time()
        # Exports the environment variables for build process.
        self.envs['COMPILE_COMPILER'] = self.option.get('compile', 'compiler')
        self.envs['COMPILE_OPTIMIZE'] = self.option.get('compile', 'optimize')
        if self.option.contains('common', 'cppflags'):
            self.envs['COMPILE_CPPFLAGS'] = \
                    self.option.get('common', 'cppflags')
        if self.option.contains('compile', 'extraflags'):
            self.envs['COMPILE_EXTRAFLAGS'] = \
                    self.option.get('compile', 'extraflags')
        # Invokes the defined tasks.
        ret = True
        self.__start_log()
        for task in self.option.getList('compile', 'task'):
            (ret, ret0) = self.__invoke(task)
            if not ret or ret0 != 0:
                ret = False
                break
        self.__stop_log()
        if not ret:
            self.__error('Failed compilation.')
        else:
            self.__info('Finished compiling the program.')
        if ret:
            self.__report('compile', 1, time.time() - t)
        else:
            self.__report('compile', 0, time.time() - t)
        return ret

    def __invoke_run(self, metric_handler):
        '''
        Invokes tasks defined for the run section.
        '''
        import os
        import time
        self.__info('Running the program.')
        os.mkdir(self.rundir)
        os.chdir(self.rundir)
        t = time.time()
        # prepares temporary files for post analysis
        outfile = os.path.join(self.rundir, 'run.out')
        errfile = os.path.join(self.rundir, 'run.err')
        out = open(outfile, 'w')
        err = open(errfile, 'w')
        self.__start_log(stdout=out, stderr=err)
        # performs the task list
        ret = True
        for task in self.option.getList('run', 'task'):
            (ret, ret0) = self.__invoke(task)
            if not ret:
                break
        self.__info('Finished running the program.')
        # performs validate process if requested
        if ret and self.option.contains('run', 'validate'):
            for task in self.option.getList('run', 'validate'):
                (ret, ret0) = self.__invoke(task)
                if not ret or ret0 != 0:
                    ret = False
                    self.__error('Validation failed')
                    break
        # closes the temporary files and redirects them to log
        self.__stop_log()
        out.close()
        err.close()
        for line in open(outfile, 'r'):
            self.stdout.write(line)
        for line in open(errfile, 'r'):
            self.stderr.write(line)
        # performs match-based validation if requested
        if ret and self.option.contains('run', 'validon'):
            ret = False
            validon = self.option.get('run', 'validon')
            for line in open(outfile, 'r'):
                if line.find(validon) > 0:
                    ret = True
                    break
            if not ret:
                self.__error('Validation failed')
        # invokes metric handler if defined.
        if metric_handler:
            metric_handler(outfile, self.metric)
            metric_handler(errfile, self.metric)
        # cleans up temporary files
        os.remove(outfile)
        os.remove(errfile)
        if ret:
            self.__report('run', 1, time.time() - t)
        else:
            self.__report('run', 0, time.time() - t)
        return ret

    def cetus_parse(self):
        '''
        Invokes cetus' parser and returns the result.
            return      (boolean result, Program IR)
        '''
        import os
        if not self.__setup():
            self.__error('Failed to set up the benchmark - check options.')
            os.chdir(self.cwd)
            return False
        (cetus_ret, cetus_program) = self.__invoke_cetus()
        if not cetus_ret:
            os.chdir(self.cwd)
            return (False, None)
        self.__info('Everything was OK')
        os.chdir(self.cwd)
        return (True, cetus_program)

    def cetus(self):
        '''
        Performs cetus task, and returns the result.
            return      True if it is successful
        '''
        import os
        if not self.__setup():
            self.__error('Failed to set up the benchmark - check options.')
            os.chdir(self.cwd)
            return False
        (cetus_ret, cetus_program) = self.__invoke_cetus()
        if not cetus_ret:
            os.chdir(self.cwd)
            return False
        self.__info('Everything was OK')
        os.chdir(self.cwd)
        return True

    def compile(self):
        '''
        Performs two tasks, cetus and compile, and returns the result.
            return      True if it is successful
        '''
        import os
        if not self.__setup():
            self.__error('Failed to set up the benchmark - check options.')
            os.chdir(self.cwd)
            return False
        (cetus_ret, cetus_program) = self.__invoke_cetus()
        if not cetus_ret:
            os.chdir(self.cwd)
            return False
        if not self.__invoke_compile():
            os.chdir(self.cwd)
            return False
        self.__info('Everything was OK')
        os.chdir(self.cwd)
        return True

    def run(self, metric_handler=None, close_log=False):
        '''
        Performs three tasks, cetus, compile, and run, and returns the result.
            return      True if it is successful
        '''
        import os
        if not self.__setup():
            self.__error('Failed to set up the benchmark - check options.')
            os.chdir(self.cwd)
            return False
        (cetus_ret, cetus_program) = self.__invoke_cetus()
        if not cetus_ret:
            os.chdir(self.cwd)
            return False
        if not self.__invoke_compile():
            os.chdir(self.cwd)
            return False
        if not self.__invoke_run(metric_handler):
            os.chdir(self.cwd)
            return False
        self.__info('Everything was OK')
        if close_log:
            self.__close_fds()
        os.chdir(self.cwd)
        return True

    def __start_log(self, stdout=None, stderr=None):
        '''
        Starts to redirect standard out/err to self.stdout/stderr.
        '''
        import sys
        sys.stdout = stdout
        sys.stderr = stderr
        if not stdout:
            sys.stdout = self.stdout
        if not stderr:
            sys.stderr = self.stderr
        return

    def __stop_log(self):
        '''
        Stops redirection triggered by __start_log().
        '''
        import sys
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        return

    def __start_logj(self):
        '''
        Starts to redirect jvm's standard out/err to self.stdout/stderr.
        '''
        import os
        from java.lang import System
        from java.io import PrintStream
        self.system_out = System.out
        self.system_err = System.err
        self.java_out = os.path.join(os.getcwd(), 'java.out')
        self.java_err = os.path.join(os.getcwd(), 'java.err')
        System.setOut(PrintStream(self.java_out))
        System.setErr(PrintStream(self.java_err))
        return

    def __stop_logj(self):
        '''
        Stops redirection triggered by __start_logj().
        '''
        import os
        from java.lang import System
        for line in open(self.java_out, 'r'):
            self.stdout.write(line)
        for line in open(self.java_err, 'r'):
            self.stderr.write(line)
        os.remove(self.java_out)
        os.remove(self.java_err)
        System.setOut(self.system_out)
        System.setErr(self.system_err)
        return

    def parse_option(self, *files):
        '''
        Reads option file(s) and stores the options as a single option object
        attached to the benchmark object.
            files       the user-provided option file(s)
        '''
        from benchmark import Option
        import os
        import ConfigParser
        opt = Option()
        self.option = opt
        self.__fill_defaults()
        for f in files:
            parser = ConfigParser.SafeConfigParser()
            parser.read(f)
            for section in parser.sections():
                for option in parser.options(section):
                    if option.startswith('+'):
                        optname = option[1:]
                    else:
                        optname = option
                    values = parser.get(section, option).split(os.linesep)
                    value0 = values[0].strip()
                    if option.startswith('+'):
                        opt.append(section, optname, value0)
                    else:
                        opt.set(section, optname, value0)
                    for value in values[1:]:
                        opt.append(section, optname, value)
        return opt

def loop_time(file, metric):
    '''
    Example metric handler to transform event timer to metric format
    TODO: This can be standardized within cetus passes
    '''
    f = open(file, 'r')
    for line in f:
        if line.startswith('CETUS_TIMING'):
            columns = line.split()
            if columns[1] != 'NAME' and \
                    columns[1] != 'PROGRAM' and \
                    columns[1] != 'EVENTSUM' and \
                    columns[1] != 'COVERAGE':
                metric.add('runtime', columns[1], columns[4])
                metric.add('invocation', columns[1], columns[2])
    f.close()
    return
