package cetus.application;

import cetus.analysis.CFGraph;
import cetus.analysis.DFAGraph;
import cetus.analysis.DFANode;
import cetus.hir.Procedure;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * This class builds the control dependence graph using the post dominance tree.
 * @author Jae-Woo Lee, <jaewoolee@purdue.edu>
 *         School of ECE, Purdue University
 */
public class ControlDependenceGraph {

    Map<Procedure, CFGraph> cfgMap;

    public void buildGraph(Map<Procedure, CFGraph> cfgMap) {
        this.cfgMap = cfgMap;
        buildCDG();
    }

    private void buildCDG() {
        Set<Procedure> procSet = cfgMap.keySet();
        for (Procedure proc : procSet) {
            CFGraph cfg = cfgMap.get(proc);
            // reverse the control flow graph
            DFAGraph reversedCFG = SlicingTools.reverseCFGraph(cfg);
            ArrayList<DFANode> nodeList = new ArrayList<DFANode>();
            
            // extract dominator
            BitSet[] dominator = SlicingTools.getDominators(reversedCFG, nodeList);
            int nodeSize = nodeList.size();
            DFANode[] nodeArray = new DFANode[nodeSize];
            nodeList.toArray(nodeArray);
            Set<Integer> entryIdxSet = SlicingTools.getEntryIdxSet(reversedCFG, nodeList);
            List<DFANode> entryNodeList = reversedCFG.getEntryNodes();
            
            // print dominator
            SlicingTools.printDominator(dominator, nodeArray, proc);
            
            // extract immediate dominator
            BitSet[] immediateDom = SlicingTools.getImmediateDominator(dominator, entryIdxSet, entryNodeList, nodeArray);
            
            // print immediate dominator
            SlicingTools.printDominator(immediateDom, nodeArray, proc);
            
            // build Post Dominance Tree
            System.out.println("############# build Post Dominace Tree (Graph) for " + proc.getSymbolName() + " ###########");
            DFAGraph postDomTree = SlicingTools.buildImmediateDominanceTree(immediateDom, nodeArray, nodeList);
            
            // print post dominance tree
            SlicingTools.printDominanceTree(postDomTree, proc);
            
            // extract edges m --> n such that n does not postdominate(dominate in reversed) m
            // CFGEdge's head and tail are on reversed cfg.
            // can get PDTree node by getData("pdtNode")
            Set<CFGEdge> nonPDEdgeSet = extractNonPostdominatingEdges(reversedCFG, postDomTree, proc);

            List<DFANode> pdtNodeList = new ArrayList<DFANode>();
            Iterator<DFANode> pdtIter = postDomTree.iterator();
            int idx = 0;
            while (pdtIter.hasNext()) {
            	DFANode node = pdtIter.next();
            	pdtNodeList.add(node);
//            	System.out.println("[" + idx++ + "]" + ((DFANode)((DFANode)node.getData("revNode")).getData("cfgNode")).getData("ir"));
            }
            for (CFGEdge edge: nonPDEdgeSet) {
                // find lowest common ancestor of m and n (result: l)
            	DFANode commonAncestor = getLowestCommonAncestor(edge, postDomTree, pdtNodeList); 

            	// all nodes in N on the path from l to n in the postdominance tree except l
                // are control dependent on m
            	Set<DFANode> cdSet = getControlDependentSet(edge, commonAncestor);
            	DFANode controllingNode = edge.tail.getData("cfgNode");
            	for (DFANode node: cdSet) {
            		DFANode revNode = node.getData("revNode");
            		if (revNode == null) {
            			continue;
            		}
            		DFANode controlledNode = revNode.getData("cfgNode");
            		controlledNode.putData("controlNode", controllingNode);
            		System.out.println("[Node]" + controlledNode.getData("ir") + " == control dependent on == [Node]" + controllingNode.getData("ir"));
            	}
            }            
        }
    }
    
    private Set<CFGEdge> extractNonPostdominatingEdges(DFAGraph reverseCFG, DFAGraph postDomTree, Procedure proc) {
    	System.out.println("[extractNonPostdominatingEdges]Proc: " + proc.getSymbolName());
    	Set<CFGEdge> retSet = new HashSet<CFGEdge>();
    	// extract all the edges from CFG
    	Set<CFGEdge> allEdgeList = new HashSet<CFGEdge>();
    	Iterator<DFANode> iter = reverseCFG.iterator();
    	while(iter.hasNext()) {
    		DFANode currentNode = iter.next();
    		Set<DFANode> predSet = currentNode.getPreds();
    		for (DFANode pred: predSet) {
    			// In the reversed cfg, pred is succ in CFG
    			allEdgeList.add(new CFGEdge(currentNode, pred));
    		}
    	}
    	List<DFANode> entryNodes = postDomTree.getEntryNodes();
    	if (entryNodes.size() != 1) {
    		throw new RuntimeException("Size of Entry node on PD Tree is not 1. Size: " + entryNodes.size());
    	}
    	DFANode entryNode = entryNodes.get(0);
    	// extract edge which head does not postdominate tail
    	// (head is not parent of tail on PostDomTree)
    	for (CFGEdge edge: allEdgeList) {
    		// 
    		DFANode pdtTail = (DFANode)edge.tail.getData("pdtNode");
    		DFANode pdtHead = (DFANode)edge.head.getData("pdtNode");
    		Set<DFANode> predSet = pdtTail.getPreds();
    		boolean postDom = false;
    		while (predSet.size() != 0) {
    			DFANode pdtNode = predSet.iterator().next();
    			if (pdtNode.equals(pdtHead)) {
    				postDom = true;
    				break;
    			}
    			predSet = pdtNode.getPreds();
    		}
    		if (postDom == false) {
    			retSet.add(edge);
    			System.out.println("Found Edge: " + ((DFANode)edge.tail.getData("cfgNode")).getData("ir") + " --> " + ((DFANode)edge.head.getData("cfgNode")).getData("ir"));
    		}
    	}
    	return retSet;
    }
    
    private DFANode getLowestCommonAncestor(CFGEdge edgeOnReversedCFG, DFAGraph postDomTree, List<DFANode> pdtNodeList) {
    	DFANode lcaNode = null;
    	String pathHead = getCommaSeparatedPath((DFANode)edgeOnReversedCFG.head.getData("pdtNode"), postDomTree, pdtNodeList);
    	String pathTail = getCommaSeparatedPath((DFANode)edgeOnReversedCFG.tail.getData("pdtNode"), postDomTree, pdtNodeList);
    	int idxCommon = SlicingTools.longestSubstr(pathHead, pathTail);
    	System.out.println("CommonString: " + pathHead.substring(0, idxCommon));
    	String commonString = pathHead.substring(0, idxCommon);
    	if (commonString.endsWith(",") == false) {
    		commonString = commonString.substring(0,commonString.lastIndexOf(","));
    	}
    	StringTokenizer tzer = new StringTokenizer(commonString, ",");
    	String lastToken = null;
    	while (tzer.hasMoreTokens()) {
    		lastToken = tzer.nextToken();
    	}
    	System.out.println("Last Token: " + lastToken);
    	lcaNode = pdtNodeList.get(Integer.parseInt(lastToken));
    	return lcaNode;
    }
    
    private String getCommaSeparatedPath(DFANode node, DFAGraph graph, List<DFANode> pdtNodeList) {
    	LinkedList<DFANode> list = new LinkedList<DFANode>();
    	list.add(node);
    	Set<DFANode> preds = node.getPreds();
    	while (preds != null && preds.size() > 0) {
    		DFANode parent = preds.iterator().next();
    		list.addFirst(parent);
    		preds = parent.getPreds();
    	}
    	StringBuffer retStr = new StringBuffer();
    	for (DFANode dfa: list) {
    		retStr.append(pdtNodeList.indexOf(dfa));
    		retStr.append(",");
    	}
    	System.out.println("String: " + retStr.toString() + ", for node: " + ((DFANode)((DFANode)node.getData("revNode")).getData("cfgNode")).getData("ir"));
    	return retStr.toString();
    }
    
    private Set<DFANode> getControlDependentSet(CFGEdge edge, DFANode commonAncestor) {
    	Set<DFANode> retSet = new HashSet<DFANode>();
    	DFANode pdtHead = edge.head.getData("pdtNode");
    	Set<DFANode> preds = pdtHead.getPreds();
    	retSet.add(pdtHead);
    	while (preds != null && preds.size() > 0) {
    		DFANode parent = preds.iterator().next();
    		if (parent.equals(commonAncestor)) {
    			break;
    		}
    		retSet.add(parent);
    		preds = parent.getPreds();
    	}
    	return retSet;
    }
}

class CFGEdge {
	DFANode tail;
	DFANode head;
	CFGEdge(DFANode tail, DFANode head) {
		this.tail = tail;
		this.head = head;
	}
}
