package cetus.application;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import cetus.analysis.CFGraph;
import cetus.analysis.DFANode;
import cetus.hir.BinaryExpression;
import cetus.hir.DoLoop;
import cetus.hir.Expression;
import cetus.hir.ForLoop;
import cetus.hir.FunctionCall;
import cetus.hir.IRTools;
import cetus.hir.Identifier;
import cetus.hir.IfStatement;
import cetus.hir.Procedure;
import cetus.hir.Program;
import cetus.hir.Statement;
import cetus.hir.Traversable;
import cetus.hir.UnaryExpression;
import cetus.hir.VariableDeclarator;
import cetus.hir.WhileLoop;

/**
 * This class slices the program using the def-use or use-def chain and
 * the control dependence graph.
 * @author Jae-Woo Lee, <jaewoolee@purdue.edu>
 *         School of ECE, Purdue University
 */
public class ProgramSlicer {
	Program program;
	Set<Procedure> procSet;
	Map<Procedure, CFGraph> cfgMap;
	UseDefChain udChain;
	
	public enum Criteria {
		PRINT_STMT
	};
	
	Map<Procedure, List<DFANode>> scMap;
	Set<DFANode> scSet;
	Set<Criteria> criteriaSet;
	boolean customCriteria = false;
	boolean predefinedCriteria = false;
	
	public ProgramSlicer(Map<Procedure, CFGraph> cfgMap, UseDefChain udChain, Program program) {
		this.cfgMap = cfgMap;
		this.program = program;
		scMap = new HashMap<Procedure, List<DFANode>>();
		scSet = new HashSet<DFANode>();
		criteriaSet = new HashSet<Criteria>();
		procSet = cfgMap.keySet();
		this.udChain = udChain;
	}
	
	public void addSlicingCriteria(DFANode dfaNode, Procedure proc) {
		customCriteria = true;
		List<DFANode> nodeList = scMap.get(proc);
		if (nodeList == null) {
			nodeList = new ArrayList<DFANode>();
			scMap.put(proc, nodeList);
		}
		nodeList.add(dfaNode);
		scSet.add(dfaNode);
	}
	
	public void addPreDefinedSlicingCriteria(Criteria cri) {
		predefinedCriteria = true;
		criteriaSet.add(cri);
	}
	
	public void markSlice() {
		if (predefinedCriteria) {
			for (Criteria cri: criteriaSet) {
				markSliceForCriteria(cri);
			}
		}
	}
	
	private void markSliceForCriteria(Criteria cri) {
		// set slice criteria
		if (cri.equals(Criteria.PRINT_STMT)) {
			System.out.println("[markSlice]mark slice criteria for print statements");
			for (Procedure proc: procSet) {
				CFGraph cfg = cfgMap.get(proc);
				Iterator<DFANode> cfgIter = cfg.iterator();
				while (cfgIter.hasNext()) {
					DFANode dfaNode = cfgIter.next();
					Traversable ir = (Traversable)CFGraph.getIR(dfaNode);
					if (IRTools.containsFunctionCall(ir)) {
						List<FunctionCall> funcList = IRTools.getFunctionCalls(ir);
						for (FunctionCall fc: funcList) {
							String fname = fc.getName().toString();
							if (fname.equals("printf")) {
								System.out.println("[markSlice]Adding slice criteria: " + ir + ", proc: " + proc.getSymbolName());
								addSlicingCriteria(dfaNode, proc);
							}
						}
					}					
				}
			}
		}
		// mark slice
		LinkedList<DFANode> workList = new LinkedList<DFANode>();
		workList.addAll(scSet);
		while (workList.isEmpty() == false) {
			DFANode currentNode = workList.remove();
//			System.out.println("[Slice]slice: " + currentNode.getData("ir"));
			currentNode.putData("isSlice", Boolean.TRUE);
			// add control dependence node
			DFANode controllingNode = currentNode.getData("controlNode");
			if (controllingNode != null&& controllingNode.getData("isSlice") == null) {
				workList.add(controllingNode);
			}
			// add def node
			Traversable currentIR = (Traversable) CFGraph.getIR(currentNode);
			List<Expression> useList = ChainTools.getUseList(currentIR);
			for (Expression use: useList) {
				Set<DFANode> defDFANodeSet = udChain.getDefDFANodeSet(use);
				for (DFANode dfaNode: defDFANodeSet) {
					if (dfaNode.getData("isSlice") == null) {
						workList.add(dfaNode);
					}
				}
			}
			// add return node of called function
			List<FunctionCall> fcList = IRTools.getFunctionCalls(currentIR);
			for (FunctionCall fc: fcList) {
				Procedure callee = fc.getProcedure();
				if (callee != null) {
					CFGraph calleeCFG = cfgMap.get(callee);
					List<DFANode> exitList = calleeCFG.getExitNodes();
					for (DFANode dfaNode: exitList) {
						if (dfaNode.getData("isSlice") == null) {
							workList.add(dfaNode);
						}
					}
				}
			}
		}
	}
	
	public void removeUnmarkedDFANode() {
		for (Procedure proc: procSet) {
			CFGraph cfg = cfgMap.get(proc);
			Iterator cfgIter = cfg.iterator();
			while (cfgIter.hasNext()) {
				DFANode dfa = (DFANode) cfgIter.next();
				if (dfa.getData("isSlice") == null) {
					Traversable tr = (Traversable) CFGraph.getIR(dfa);
					if (tr == null) {
						continue;
					}
					if (tr instanceof Statement) {
						Statement ex = (Statement) tr;
						Traversable stmtParent = ex.getParent();
						if (stmtParent instanceof ForLoop) { 
							Traversable stmtParParent = stmtParent.getParent();
							stmtParParent.removeChild(stmtParent);
						} else {
							stmtParent.removeChild(ex);
						}
					} else if (tr instanceof BinaryExpression) {
						// This handles IfStatement
						BinaryExpression ex = (BinaryExpression) tr;
						Traversable binaryParent = ex.getParent();
						if (binaryParent instanceof IfStatement) {
							Traversable binaryParParent = binaryParent.getParent();
							binaryParParent.removeChild(binaryParent);
						} else if (binaryParent instanceof ForLoop) {
							Traversable stmtParParent = binaryParent.getParent();
							stmtParParent.removeChild(binaryParent);
						} else if (binaryParent instanceof WhileLoop) {
							Traversable stmtParParent = binaryParent.getParent();
							stmtParParent.removeChild(binaryParent);
						} else if (binaryParent instanceof DoLoop) {
							Traversable stmtParParent = binaryParent.getParent();
							stmtParParent.removeChild(binaryParent);
						} else {
							throw new RuntimeException("[removeUnmarkedDFANode]Unhandled Type: " + tr.toString() + ", proc: " + proc.getSymbolName() + ", class: " + tr.getClass().getCanonicalName());
						}
					} else if (tr instanceof VariableDeclarator) {
						VariableDeclarator vDeclarator = (VariableDeclarator) tr;
						Traversable parentTrav = vDeclarator.getParent();
						// The following operation is not allowed.
						// parentTrav.removeChild(vDeclarator);
					} else if (tr instanceof UnaryExpression) {
						// the removing is not allowed
					} else if (tr instanceof Identifier) {
						// the removing is not allowed
					} else if (tr instanceof FunctionCall) {
						Traversable fcParent = tr.getParent();
						if (fcParent instanceof IfStatement) {
							Traversable fcParParent = fcParent.getParent();
							fcParParent.removeChild(fcParent);
						} else {
							throw new RuntimeException("[removeUnmarkedDFANode]Unhandled Type: " + tr.toString() + ", proc: " + proc.getSymbolName() + ", class: " + tr.getClass().getCanonicalName());
						}
					} else {
						throw new RuntimeException("[removeUnmarkedDFANode]Unhandled Type: " + tr.toString() + ", proc: " + proc.getSymbolName() + ", class: " + tr.getClass().getCanonicalName());
					}
				}
			}
		}
	}
	
	public void writeToDisk() {
		try {
			program.print();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
}
